@@ -61,8 +61,8 @@ mod ir_builder {
                 }
 
             rule fn_local() -> (IrAstTy, String, bool, Option<IrAstOperation>)
-                = "local" _ im:("mut" _)? "ptr" _ ty:ast_ty() name:id() init:fn_local_init()? {
-                    (ty, name, im.is_some(), init)
+                = "local" _ im:mut_ptr() ty:ast_ty() name:id() init:fn_local_init()? {
+                    (ty, name, im, init)
                 }
 
             rule fn_local_init() -> IrAstOperation
@@ -114,8 +114,12 @@ mod ir_builder {
                 / op_insert_element()
                 / op_insert_value()
                 / op_load()
+                / op_nop()
                 / op_phi()
+                / op_ptr_cast()
                 / op_ret()
+                / op_state_load()
+                / op_state_store()
                 / op_store()
 
             rule op_asm() -> IrAstOperation
@@ -156,7 +160,7 @@ mod ir_builder {
                 }
 
             rule op_get_ptr() -> IrAstOperation
-                = "get_ptr" _ ("mut" _)? "ptr" _ ty:ast_ty() name:id() {
+                = "get_ptr" _ mut_ptr() ty:ast_ty() name:id() {
                     IrAstOperation::GetPtr(name)
                 }
 
@@ -171,23 +175,43 @@ mod ir_builder {
                 }
 
             rule op_load() -> IrAstOperation
-                = "load" _ ("mut" _)? "ptr" _ ast_ty() src:id() {
+                = "load" _ ptr() src:id() {
                     IrAstOperation::Load(src)
                 }
 
+            rule op_nop() -> IrAstOperation
+                = "nop" _ {
+                    IrAstOperation::Nop
+                }
+
             rule op_phi() -> IrAstOperation
                 = "phi" _ "(" _ pairs:((bl:id() ":" _ vn:id() { (bl, vn) }) ** comma()) ")" _ {
                     IrAstOperation::Phi(pairs)
                 }
 
+            rule op_ptr_cast() -> IrAstOperation
+                = "ptr_cast" _ ptr() vn:id() "to" _ ptr() ty:ast_ty() {
+                    IrAstOperation::PtrCast(vn, ty)
+            }
+
             rule op_ret() -> IrAstOperation
                 = "ret" _ ty:ast_ty() vn:id() {
                     IrAstOperation::Ret(ty, vn)
                 }
 
+            rule op_state_load() -> IrAstOperation
+                = "state_load" _ ptr() dst:id() comma() "key" _ key:id() {
+                    IrAstOperation::StateLoad(dst, key)
+                }
+
+            rule op_state_store() -> IrAstOperation
+                = "state_store" _ ptr() src:id() comma() "key" _ key:id() {
+                    IrAstOperation::StateStore(src, key)
+                }
+
             rule op_store() -> IrAstOperation
-                = "store" _ dst:id() comma() ("mut" _)? "ptr" _ ast_ty() vn:id() {
-                    IrAstOperation::Store(dst, vn)
+                = "store" _ val:id() comma() ptr() dst:id() {
+                    IrAstOperation::Store(val, dst)
                 }
 
             rule asm_arg() -> (Ident, Option<IrAstAsmArgInit>)
@@ -350,6 +374,14 @@ mod ir_builder {
                     ds.parse::<u64>().unwrap()
                 }
 
+            rule ptr()
+                = "ptr" _
+
+            rule mut_ptr() -> bool
+                = m:("mut" _)? ptr() {
+                    m.is_some()
+                }
+
             rule comma()
                 = quiet!{ "," _ }
 
@@ -380,6 +412,7 @@ mod ir_builder {
         context::Context,
         error::IrError,
         function::Function,
+        instruction::Instruction,
         irtype::{Aggregate, Type},
         metadata::{MetadataIndex, Metadatum},
         module::{Kind, Module},
@@ -434,8 +467,12 @@ mod ir_builder {
         InsertElement(String, IrAstTy, String, String),
         InsertValue(String, IrAstTy, String, Vec<u64>),
         Load(String),
+        Nop,
         Phi(Vec<(String, String)>),
+        PtrCast(String, IrAstTy),
         Ret(IrAstTy, String),
+        StateLoad(String, String),
+        StateStore(String, String),
         Store(String, String),
     }
 
@@ -582,17 +619,21 @@ mod ir_builder {
         let mut ctx = Context::default();
         let module = Module::new(&mut ctx, ir_ast_mod.kind);
         let md_map = build_metadata_map(&mut ctx, &ir_ast_mod.metadata);
+        let mut unresolved_calls = Vec::new();
         for fn_decl in ir_ast_mod.fn_decls {
-            build_add_fn_decl(&mut ctx, module, fn_decl, &md_map)?;
+            build_add_fn_decl(&mut ctx, module, fn_decl, &md_map, &mut unresolved_calls)?;
         }
+        resolve_calls(&mut ctx, unresolved_calls)?;
         Ok(ctx)
     }
 
+    #[allow(clippy::type_complexity)]
     fn build_add_fn_decl(
         context: &mut Context,
         module: Module,
         fn_decl: IrAstFnDecl,
         md_map: &HashMap<MdIdxRef, MetadataIndex>,
+        unresolved_calls: &mut Vec<(Block, Value, String, Vec<Value>, Option<MetadataIndex>)>,
     ) -> Result<(), IrError> {
         let args: Vec<(String, Type, Option<MetadataIndex>)> = fn_decl
             .args
@@ -658,11 +699,13 @@ mod ir_builder {
                 &ptr_map,
                 &mut arg_map,
                 md_map,
+                unresolved_calls,
             );
         }
         Ok(())
     }
 
+    #[allow(clippy::type_complexity)]
     fn build_add_block_instructions(
         context: &mut Context,
         ir_block: IrAstBlock,
@@ -670,6 +713,7 @@ mod ir_builder {
         ptr_map: &HashMap<String, Pointer>,
         val_map: &mut HashMap<String, Value>,
         md_map: &HashMap<MdIdxRef, MetadataIndex>,
+        unresolved_calls: &mut Vec<(Block, Value, String, Vec<Value>, Option<MetadataIndex>)>,
     ) {
         let block = named_blocks.get(&ir_block.label).unwrap();
         for ins in ir_block.instructions {
@@ -703,8 +747,7 @@ mod ir_builder {
                                 immediate: imm,
                                 span_md_idx: meta_idx
                                     .as_ref()
-                                    .map(|meta_idx| md_map.get(meta_idx).copied())
-                                    .flatten(),
+                                    .and_then(|meta_idx| md_map.get(meta_idx).copied()),
                             },
                         )
                         .collect();
@@ -718,26 +761,21 @@ mod ir_builder {
                     block.ins(context).branch(*to_block, None, opt_ins_md_idx)
                 }
                 IrAstOperation::Call(callee, args) => {
-                    let function = context
-                        .functions
-                        .iter()
-                        .find_map(|(idx, content)| {
-                            if content.name == callee {
-                                Some(Function(idx))
-                            } else {
-                                None
-                            }
-                        })
-                        .unwrap();
-                    block.ins(context).call(
-                        function,
-                        &args
-                            .iter()
+                    // We can't resolve calls to other functions until we've done a first pass and
+                    // created them first.  So we can insert a NOP here, save the call params and
+                    // replace it with a CALL in a second pass.
+                    let nop = block.ins(context).nop();
+                    unresolved_calls.push((
+                        *block,
+                        nop,
+                        callee,
+                        args.iter()
                             .map(|arg_name| val_map.get(arg_name).unwrap())
                             .cloned()
                             .collect::<Vec<Value>>(),
                         opt_ins_md_idx,
-                    )
+                    ));
+                    nop
                 }
                 IrAstOperation::Cbr(cond_val_name, true_block_name, false_block_name) => {
                     block.ins(context).conditional_branch(
@@ -792,7 +830,8 @@ mod ir_builder {
                 }
                 IrAstOperation::Load(src_name) => block
                     .ins(context)
-                    .load(*ptr_map.get(&src_name).unwrap(), opt_ins_md_idx),
+                    .load(*val_map.get(&src_name).unwrap(), opt_ins_md_idx),
+                IrAstOperation::Nop => block.ins(context).nop(),
                 IrAstOperation::Phi(pairs) => {
                     for (block_name, val_name) in pairs {
                         block.add_phi(
@@ -803,14 +842,32 @@ mod ir_builder {
                     }
                     block.get_phi(context)
                 }
+                IrAstOperation::PtrCast(value_name, ty) => {
+                    let ty = ty.to_ir_type(context);
+                    block.ins(context).ptr_cast(
+                        *val_map.get(&value_name).unwrap(),
+                        ty,
+                        opt_ins_md_idx,
+                    )
+                }
                 IrAstOperation::Ret(ty, ret_val_name) => {
                     let ty = ty.to_ir_type(context);
                     block
                         .ins(context)
                         .ret(*val_map.get(&ret_val_name).unwrap(), ty, opt_ins_md_idx)
                 }
-                IrAstOperation::Store(stored_val_name, ptr_name) => block.ins(context).store(
-                    *ptr_map.get(&ptr_name).unwrap(),
+                IrAstOperation::StateLoad(dst, key) => block.ins(context).state_load(
+                    *val_map.get(&dst).unwrap(),
+                    *val_map.get(&key).unwrap(),
+                    opt_ins_md_idx,
+                ),
+                IrAstOperation::StateStore(src, key) => block.ins(context).state_store(
+                    *val_map.get(&src).unwrap(),
+                    *val_map.get(&key).unwrap(),
+                    opt_ins_md_idx,
+                ),
+                IrAstOperation::Store(stored_val_name, dst_val_name) => block.ins(context).store(
+                    *val_map.get(&dst_val_name).unwrap(),
                     *val_map.get(&stored_val_name).unwrap(),
                     opt_ins_md_idx,
                 ),
@@ -856,6 +913,34 @@ mod ir_builder {
         }
         md_map
     }
+
+    #[allow(clippy::type_complexity)]
+    fn resolve_calls(
+        context: &mut Context,
+        unresolved_calls: Vec<(Block, Value, String, Vec<Value>, Option<MetadataIndex>)>,
+    ) -> Result<(), IrError> {
+        // All of the call instructions are currently NOPs which need to be replaced with actual
+        // calls.  We couldn't do it above until we'd gone and created all the functions first.
+        //
+        // Now we can loop and find the callee function for each call and replace the NOPs.
+        for (block, nop, callee, args, opt_ins_md_idx) in unresolved_calls {
+            let function = context
+                .functions
+                .iter()
+                .find_map(|(idx, content)| {
+                    if content.name == callee {
+                        Some(Function(idx))
+                    } else {
+                        None
+                    }
+                })
+                .unwrap();
+            let call_val =
+                Value::new_instruction(context, Instruction::Call(function, args), opt_ins_md_idx);
+            block.replace_instruction(context, nop, call_val)?;
+        }
+        Ok(())
+    }
 }
 
 // -------------------------------------------------------------------------------------------------
